"""QtFrameRate widget.

Displays the framerate as colored LEDs.
"""
import math
import time
from typing import Optional

import numpy as np
from qtpy.QtCore import QTimer
from qtpy.QtGui import QImage, QPixmap
from qtpy.QtWidgets import QLabel

# Shape of the bitmap.
BITMAP_SHAPE = (20, 270, 4)

# Number of left to right LED segments.
NUM_SEGMENTS = 30

# Spacing between segments, including the gap.
SEGMENT_SPACING = BITMAP_SHAPE[1] / NUM_SEGMENTS

# Gap between segments in pixels.
SEGMENT_GAP = 2
SEGMENT_WIDTH = SEGMENT_SPACING - SEGMENT_GAP

# DECAY_MS (X, Y) means if segment is at least X then keep that LED
# on for Y milliseconds. Peak means we hit that exact value,
# while active means it was a leading LED on the way to that peak.
#
# The general goals for the LEDs are:
#
# 1) The current peak is bright and easily visible.
# 2) The LEDs leading up to the current peak are bright so that makes
#    the current peak easier to see.
# 3) The peaks decay more slowly than the leading LEDs. So the slow frames
#    are visible even when the framerate has improved. So we can
#    see the bad peaks for a while.
#
# Make yellow and green stay lit longer, so we don't miss them.
DECAY_MS = {"peak": [(0, 1000), (10, 2000), (20, 3000)], "active": [(0, 250)]}

# LEDs are colors GREEN -> YELLOW -> RED using slightly non-pure colors.
# Colors were chosen quickly and could be tweaked.
GREEN = (57, 252, 3)
YELLOW = (252, 232, 3)
RED = (252, 78, 3)

# Ten segments of each color.
LED_COLORS = [(0, GREEN), (10, YELLOW), (20, RED)]

# LEDs decay from MAX to OFF. We leave them visible when off because it
# looks better, and so you can see how high the meter goes.
ALPHA_MAX = 255
ALPHA_OFF = 25

# We should have better knobs for calibrating which framerates lead to
# which colors. But playing around this values we now get:
#
# GREEN: up to 31ms (32Hz)
# YELLOW: up to 312ms (3.1Hz)
# RED: up to 6000ms (0.16Hz)
#
# That seems pretty useful. Green is good, yellow is bad, red is awful.
#
PERFECT_MS = 16.7
LOG_BASE = 1.35


def _clamp(value, low, high):
    return max(min(value, high), low)


def _lerp(low: int, high: int, fraction: float) -> int:
    return int(low + (high - low) * fraction)


def _decay_seconds(index: int, decay_config) -> float:
    """Return duration in seconds the segment should stay on.

    Only called during init to create a lookup table, for speed.

    Parameters
    ----------
    index : int
        The LED segment index.

    Return
    ------
    float
        Time in milliseconds to stay lit up.
    """
    for limit, ms in reversed(decay_config):
        if index >= limit:
            return ms / 1000
    assert False, "DECAY_MS table is messed up?"
    return 0


def _led_color(index: int) -> tuple:
    """Return color of the given LED segment index.

    Only called during init to create a lookup table, for speed.

    Parameters
    ----------
    index : int
        The led segment index.
    """
    for limit, color in reversed(LED_COLORS):
        if index >= limit:
            return color
    assert False, "DECAY_MS table is messed up?"
    return (0, 0, 0)


def _get_peak(delta_seconds: float) -> int:
    """Get highest segment that should be lit up.

    Parameters
    ----------
    delta_seconds : float
        The current frame interval.

    Return
    ------
    int
        The peak segment that should be lit up.
    """

    if delta_seconds <= 0:
        return 0

    # Slide everything from PERFECT_MS on down to zero, or else we'd
    # waste a lot of LEDs show stuff faster than 60Hz.
    input_value = max((delta_seconds * 1000) - PERFECT_MS, 0) + 1

    log_value = math.log(input_value, LOG_BASE)
    return _clamp(int(log_value), 0, NUM_SEGMENTS - 1)


class DecayTimes:
    """Keep track of when LEDs were last lit up.

    Parameters
    ----------
    config : str
        Config is 'active' or 'peak'.
    """

    def __init__(self, config: str):
        count = NUM_SEGMENTS
        decay_ms = DECAY_MS[config]
        self._decay = np.array(
            [_decay_seconds(i, decay_ms) for i in range(count)], dtype=np.float
        )
        self._last = np.zeros((count), dtype=np.float)

    def is_off(self, index: int) -> bool:
        """Return True if this LED is off.

        Return
        ------
        bool
            True if this LED is off.
        """
        return self._last[index] == 0

    def all_off(self) -> bool:
        """Return True if all LEDs are off.

        Return
        ------
        bool
            True if all LEDs are off.
        """
        return np.all(self._last == 0)

    def mark(self, slicer: slice, now: float) -> None:
        """Mark that one or more LEDs have lit up.

        Parameters
        ----------
        slicer : slice
            The LEDs to light up.
        now : float
            The current time in seconds.
        """
        self._last[slicer] = now

    def expire(self, now: float) -> None:
        """Zero out the last time of any expired LEDs.

        Parameters
        ----------
        now : float
            The current time in seconds.
        """
        # Set self._last equal to zero for any LED which as gone dark.
        elapsed = now - self._last
        self._last[elapsed > self._decay] = 0

    def get_alpha(self, now: float) -> np.ndarray:
        """Return alpha for every LED segment.

        Alpha values will range from ALPHA_MAX if the LED was recently lit
        up to ALPHA_OFF if the LED is past the its decay period.

        Parameters
        ----------
        now : float
            The current time in seconds.

        Return
        ------
        np.ndarray
            The alpha values for each LED.
        """
        fractions = (now - self._last) / self._decay
        return np.interp(fractions, [0, 1], [ALPHA_MAX, ALPHA_OFF])


def _print_calibration():
    """Print mapping from LED segment to millseconds.

    Since we don't yet have the reverse math figured out, we print out
    the mapping by sampling. Crude but works.
    """
    print(f"LOG_BASE = {LOG_BASE}")
    previous = None
    for ms in range(0, 10000):
        segment = _get_peak(ms / 1000)
        if previous != segment:
            print(f"LED: {segment} -> {ms}ms")
            previous = segment


class LedState:
    """State of the LEDs in the frame rate display."""

    def __init__(self):

        # Active means the LED lit up because it was leading LED before the
        # peak value. While peaks means the framerate was that exact value.
        self._active = DecayTimes('active')
        self._peak = DecayTimes('peak')

        # Color at each LED index.
        self._color = np.array(
            [_led_color(i) for i in range(NUM_SEGMENTS)], dtype=np.uint8
        )

        self.peak = 0  # Current highest segment lit up

    def all_off(self) -> bool:
        """Return True if all LEDs are off.

        Return
        ------
        bool
            True if all LEDs are off.
        """
        return self._active.all_off() and self._peak.all_off()

    def set_peak(self, now: float, delta_seconds: float) -> None:
        """Set the peak LED based on this frame time.

        Parameters
        ----------
        now : float
            Current time in seconds.
        delta_seconds : float
            The last frame time.
        """
        # Compute the peak LED segment that should light up. Note this is
        # logarithmic so the red LEDs are much slower frames than green.
        self.peak = _get_peak(delta_seconds)

        # Mark everything before the peak as active. These light up and
        # fade but not as bright. So we can see the peaks.
        up_to_peak = slice(0, self.peak + 1)
        self._active.mark(up_to_peak, now)

        # This specific LED gets fully lit up. That way we can see this
        # peak even when the leading LEDs have faded.
        only_peak = slice(self.peak, self.peak + 1)
        self._peak.mark(only_peak, now)

    def update(self, now: float) -> None:
        """Update the LED states.

        Parameters
        ----------
        now : float
            Current time in seconds.
        """
        self._active.expire(now)
        self._peak.expire(now)

    def get_colors(self, now: float) -> np.ndarray:
        """Get current color (with alpha for our LED segments.

        Parameters
        ----------
        now : float
            The current time in seconds.

        Return
        ------
        np.ndarray
            Color (r, g, b, a)values for each LED.
        """
        # Use whichever alpha is higher (brighter).
        alpha = np.maximum(
            self._active.get_alpha(now), self._peak.get_alpha(now)
        )

        # Combine the fixed colors with the computed alpha values.
        return np.hstack((self._color, alpha.reshape(-1, 1)))


class QtFrameRate(QLabel):
    """A frame rate label with green/yellow/red LEDs.

    The LED values are logarithmic, so a lot of detail between 60Hz and
    10Hz but then the highest LED is many seconds long.
    """

    def __init__(self):
        super().__init__()

        self.leds = LedState()  # The per-LED config and state.

        # The last time we were updated, either from mouse move or our
        # timer. We update _last_time in both cases.
        self._last_time: Optional[float] = None

        # The bitmap image we draw into.
        self._image = np.zeros(BITMAP_SHAPE, dtype=np.uint8)

        # We animate on camera movements, but then we use a timer to
        # animate the display. When all the LEDs go off, we stop the timer
        # so that we use zero CPU until another camera movement.
        self._timer = QTimer()
        self._timer.setSingleShot(False)
        self._timer.setInterval(33)
        self._timer.timeout.connect(self._on_timer)

        # _print_calibration()  # Debugging.

    def _on_timer(self) -> None:
        """Animate the LEDs."""
        now = time.time()
        self._draw(now)  # Just animate and draw, no new peak.

        # Stop timer if nothing more to animation, save CPU.
        if self.leds.all_off():
            self._timer.stop()

    def _draw(self, now: float) -> None:
        """Animate the LEDs.

        Parameters
        ----------
        now : float
            The current time in seconds.
        """
        self.leds.update(now)  # Animates the LEDs.
        self._update_image(now)  # Draws our internal self._image
        self._update_bitmap()  # Writes self._image into the QLabel bitmap.

        # We always update _last_time whether this was from a camera move
        # or the timer. This is the right thing to do since in either case
        # it means a frame was drawn.
        self._last_time = now

    def on_camera_move(self) -> None:
        """Update our display to show the new framerate."""

        # Only count this frame if the timer is active. This avoids
        # displaying a potentially super long frame since it might have
        # been many seconds or minutes since the last camera movement.
        #
        # Ideally we should display the draw time of even that first frame,
        # but there's no easy/obvious way to do that today. And this will
        # show everthing except one frame.
        first_time = self._last_time is None
        use_delta = self._timer.isActive() and not first_time
        now = time.time()

        if use_delta:
            delta_seconds = now - self._last_time
            self.leds.set_peak(now, delta_seconds)

        self._draw(now)  # Draw the whole meter.

        # Since there was activity, we need to start the timer so we can
        # animate the decay of the LEDs. The timer will be shut off when
        # all the LEDs go idle.
        self._timer.start()

    def _update_image(self, now: float) -> None:
        """Update our self._image with the latest meter display.

        Parameters
        ----------
        now : float
            The current time in seconds.
        """
        self._image.fill(0)  # Start fresh each time.

        # Get colors with latest alpha values accord to decay.
        colors = self.leds.get_colors(now)

        # Draw each segment with the right color and alpha (due to decay).
        for index in range(NUM_SEGMENTS):
            x0 = int(index * SEGMENT_SPACING)
            x1 = int(x0 + SEGMENT_WIDTH)
            y0, y1 = 0, BITMAP_SHAPE[0]  # The whole height of the bitmap.
            self._image[y0:y1, x0:x1] = colors[index]

    def _update_bitmap(self) -> None:
        """Update the bitmap with latest image data."""
        height, width = BITMAP_SHAPE[:2]
        image = QImage(self._image, width, height, QImage.Format_RGBA8888)
        self.setPixmap(QPixmap.fromImage(image))



"""Octree class.
"""
from typing import List

from ...._vendor.experimental.humanize.src.humanize import intword
from ....utils.perf import block_timer
from .octree_level import OctreeLevel
from .octree_tile_builder import create_downsampled_levels
from .octree_util import SliceConfig


def _dim_str(dim: tuple) -> None:
    return f"{dim[0]} x {dim[1]} = {intword(dim[0] * dim[1])}"


def _print_levels(
    label: str, levels: List[OctreeLevel], start: int = 0
) -> None:
    print(f"{label} {len(levels)} levels:")
    for i, level in enumerate(levels):
        image_str = _dim_str(level.info.image_shape)
        tiles_str = _dim_str(level.info.shape_in_tiles)
        level = start + i
        print(f"    Level {level}: {image_str} pixels -> {tiles_str} tiles")


class Octree:
    """A region octree that holds hold 2D or 3D images.

    Today the octree is full/complete meaning every node has 4 or 8
    children, and every leaf node is at the same level of the tree. This
    makes sense for region/image trees, because the image exists
    everywhere.

    Since we are a complete tree we don't need actual nodes with references
    to the node's children. Instead, every level is just an array, and
    going from parent to child or child to parent is trivial, you just
    need to double or half the indexes.

    Future Work: Geometry
    ---------------------
    Eventually we want our octree to hold geometry, not just images.
    Geometry such as points and meshes. For geometry a sparse octree might
    make more sense than this full/complete region octree.

    With geometry there might be lots of empty space in between small dense
    pockets of geometry. Some parts of tree might need to be very deep, but
    it would be a waste for the tree to be that deep everywhere.

    Parameters
    ----------
    slice_id : int:
        The id of the slice this octree is in.
    base_shape : Tuple[int, int]
        The shape of the full base image.
    levels : Levels
        All the levels of the tree.
    """

    def __init__(self, slice_id: int, data, slice_config: SliceConfig):
        self.data = data
        self.slice_config = slice_config

        self.levels = [
            OctreeLevel(slice_id, data[i], slice_config, i)
            for i in range(len(data))
        ]

        if not self.levels:
            # Probably we will allow empty trees, but for now raise:
            raise ValueError(
                f"Data of shape {data.shape} resulted " "no octree levels?"
            )

        _print_levels("Octree input data has", self.levels)
        original_levels = len(self.levels)

        # If root level contains more than one tile, add more levels
        # until the root does consist of a single tile.
        if self.levels[-1].info.num_tiles > 1:
            with block_timer("_create_additional_levels") as timer:
                more_levels = self._create_additional_levels(slice_id)

            _print_levels(
                f"In {timer.duration_ms:.3f}ms created",
                more_levels,
                start=original_levels,
            )
            self.levels.extend(more_levels)

            print(f"Tree now has {len(self.levels)} total levels.")

        # Now the root should definitely contain only a single tile.
        assert self.levels[-1].info.num_tiles == 1

        # This now the total number of levels.
        self.num_levels = len(data)

    def _create_additional_levels(self, slice_id: int) -> List[OctreeLevel]:
        """Add additional levels to the octree.

        Keep adding levels until we each a root level where the image
        data fits inside a single tile.

        Parameters
        -----------
        slice_id : int
            The id of the slice this octree is in.
        tile_size : int
            Keep creating levels until one fits with a tile of this size.

        Return
        ------
        List[OctreeLevels]
            The new downsampled levels we created.

        Notes
        -----
        If we created this octree data, the root/highest level would
        consist of a single tile. However, if we are reading someone
        else's multiscale data, they did not know our tile size. Or they
        might not have imagined someone using tiled rendering at all.
        So their root level might be pretty large. We've seen root
        levels larger than 8000x8000 pixels.

        It would be nice in this case if our root level could use a
        really large tile size as a special case. So small tiles for most
        levels, but then one big tile as the root.

        However, today our TiledImageVisual can't handle that. It can't
        handle a mix of tile sizes, and TextureAtlas2D can't even
        allocate multiple textures!

        So for now, we compute/downsample additional levels ourselves and
        add them to the data. We do this until we reach a level that does
        fit within a single tile.

        For example for that 8000x8000 pixel root level, if we are using
        256x256 tiles we'll keep adding levels until we get to a level
        that fits within 256x256.

        Unfortunately, we can't be sure our downsampling approach will
        match the rest of the data. The levels we add might be more/less
        blurry than the original ones, or have different downsampling
        artifacts. That's probably okay because these are low-resolution
        levels, but still it would be better to show their root level
        verbatim on a single large tiles.

        Also, this is slow due to the downsampling, but also slow due to
        having to load the full root level into memory. If we had a root
        level that was tile sized, then we could load that very quickly.
        That issue does not go away even if we have a special large-size
        root tile. For best performance multiscale data should be built
        down to a very small root tile, the extra storage is negligible.
        """

        # Create additional data levels so that the root level
        # consists of only a single tile, using our standard/only
        # tile size.
        tile_size = self.slice_config.tile_size
        new_levels = create_downsampled_levels(self.data[-1], tile_size)

        # Add the data.
        self.data.extend(new_levels)

        # Return an OctreeLevel for each new data level.
        num_current = len(self.levels)
        return [
            OctreeLevel(
                slice_id, new_data, self.slice_config, num_current + index,
            )
            for index, new_data in enumerate(new_levels)
        ]

    def print_info(self):
        """Print information about our tiles."""
        for level in self.levels:
            level.print_info()



"""OctreeChunk class
"""
from typing import NamedTuple, Optional, Tuple

import numpy as np

from ....components.experimental.chunk import ChunkKey
from ....layers import Layer
from ....types import ArrayLike


class OctreeChunkGeom(NamedTuple):
    """Position and scale of the chunk, for rendering.

    Stored in the OctreeChunk so that we calculate them just once
    at OctreeChunk creation time.
    """

    pos: np.ndarray
    scale: np.ndarray


class OctreeLocation(NamedTuple):
    """Location of one chunk within the octree.

    This is used as part of the OctreeChunkKey to uniquely identify a
    chunk. The OctreeChunkKey is used when we load chunk and used
    related to the cache.
    """

    slice_id: int
    level_index: int
    row: int
    col: int

    def __str__(self):
        return (
            f"location=({self.level_index}, {self.row}, {self.col}) "
            f"slice={self.slice_id} id={id(self)}"
        )

    @classmethod
    def create_null(cls):
        """Create null location that points to nothing."""
        return cls(0, 0, 0, 0, np.zeros(0), np.zeros(0))


class OctreeChunk:
    """A geographically meaningful portion of the full 2D or 3D image.

    For 2D images a chunk is a "tile". It's a 2D square region of pixels
    which are part of the full 2D image. If it's in level 0 of the octree,
    the pixels are 1:1 identical to the portion of the full image. The tile
    is full resolution. If it's in level 1 or greater the pixels are
    downsampled from the full resolution image.

    For 3D, not yet implemented, a chunk is a sub-volume. Again for level 0
    the voxels are at the full resolution of the full image, but for other
    levels the voxels are downsampled.

    The highest level of the tree contains a single chunk which depicts the
    entire image, whether 2D or 3D.

    Attributes
    ----------
    data : ArrayLike
        The data to draw for this chunk.
    _orig_data : ArrayLike
        The original unloaded data that we use to implement OctreeChunk.clear().
    location : OctreeLocation
        The location of this chunk, including the level_index, row, col.
    geom : OctreeChunkGeom
        The x, y coordinates and scale of the chunk.
    loading : bool
        If True the chunk has been queued to be loaded.
    """

    def __init__(
        self, data: ArrayLike, location: OctreeLocation, geom: OctreeChunkGeom
    ):
        self._data = data
        self._orig_data = data  # For now hold on to implement clear()
        self.location = location
        self.geom = geom
        self.loading = False

    def __str__(self):
        return f"{self.location}"

    @property
    def data(self) -> ArrayLike:
        """Return the data associated with this chunk.

        Because the chunk has been loaded this might be an ndarray or it
        might be Dask array or other array-like object. After the chunk has
        been loaded it will always be an ndarray. The bytes will be
        in memory and ready to be drawn.
        """
        return self._data

    @data.setter
    def data(self, data: np.ndarray) -> None:
        """Set the new data for this chunk.

        We set the data after a chunk as been loaded..

        Parameters
        ----------
        data : np.ndarray
            The new data for the chunk.
        """
        # An ndarray mean it's actual bytes in memory.
        assert isinstance(data, np.ndarray)

        # Assign and say this in-progress load is now finished.
        self._data = data
        self.loading = False

    @property
    def key(self) -> Tuple[int, int, int]:
        """The unique key for this chunk.

        TODO_OCTREE: Switch to __hash__? Tried __hash__ a while ago and ran
        into problems, but maybe try again.
        """
        return (
            self.geom.pos[0],
            self.geom.pos[1],
            self.location.level_index,
        )

    @property
    def in_memory(self) -> bool:
        """Return True if the data is fully in memory.

        Return
        ------
        bool
            True if data is fully in memory.
        """
        return isinstance(self.data, np.ndarray)

    @property
    def needs_load(self) -> bool:
        """Return true if this chunk needs to loaded.

        An unloaded chunk's data might be a Dask or similar deferred array.
        A loaded chunk's data is always ndarray.

        Return
        ------
            True if the chunk needs to be loaded.
        """
        return not self.in_memory and not self.loading

    def clear(self) -> None:
        """Clear out our loaded data, return to the original.

        This is only done when running without the cache, so that we reload
        the data again. With computation the loaded data might be different
        each time, so we need to do it each time.

        TODO_OCTREE: Depending on how we end up doing caching we might
        no longer need this method?
        """
        self._data = self._orig_data
        self.loading = False


class OctreeChunkKey(ChunkKey):
    """A ChunkKey with octree specific fields.

    Parameters
    ----------
    layer : Layer
        The OctreeImage layer.
    indices : Tuple[Optional[slice], ...]
        The indices of the image we are viewing.
    location : OctreeLocation
        The location of the chunk within the octree.
    """

    def __init__(
        self,
        layer: Layer,
        indices: Tuple[Optional[slice], ...],
        location: OctreeLocation,
    ):
        self.location = location
        super().__init__(layer, indices)

    def _get_hash_values(self):
        # TODO_OCTREE: can't we just has with parent's hashed key instead
        # of creating a single big has value? Probably.
        parent = super()._get_hash_values()
        return parent + (self.location,)



"""OctreeLevel and OctreeLevelInfo classes.
"""
import math
from typing import Optional

import numpy as np

from ....types import ArrayLike
from .octree_chunk import OctreeChunk, OctreeChunkGeom, OctreeLocation
from .octree_util import SliceConfig


class OctreeLevelInfo:
    """Information about one level of the octree.

    Parameters
    ----------
    slice_config : SliceConfig
        Information about the entire octree.
    level_index : int
        The index of this level within the whole tree.
    shape_in_tiles : Tuple[int, int]
        The (height, width) dimensions of this level in terms of tiles.
    """

    def __init__(self, slice_config: SliceConfig, level_index: int):
        self.slice_config = slice_config

        self.level_index = level_index
        self.scale = 2 ** self.level_index

        base = slice_config.base_shape
        self.image_shape = (
            int(base[0] / self.scale),
            int(base[1] / self.scale),
        )

        tile_size = self.slice_config.tile_size
        scaled_size = tile_size * self.scale

        self.shape_in_tiles = [
            math.ceil(base[0] / scaled_size),
            math.ceil(base[1] / scaled_size),
        ]

        self.num_tiles = self.shape_in_tiles[0] * self.shape_in_tiles[1]


class OctreeLevel:
    """One level of the octree.

    A level contains a 2D array of tiles.

    Soon might also contain a 3D array of sub-volumes.

    Parameters
    ----------
    slice_id : int
        The id of the OctreeMultiscaleSlice we are in.
    data : ArrayLike
        The data for this level.
    slice_config : SliceConfig
        The base image shape and other details.
    level_index : int
        Index of this specific level (0 is full resolution).
    """

    def __init__(
        self,
        slice_id: int,
        data: ArrayLike,
        slice_config: SliceConfig,
        level_index: int,
    ):
        self.slice_id = slice_id
        self.data = data

        # TODO_OCTREE: change from "info" to "meta"/"metadata"?
        # info is kind of dumb sounding.
        self.info = OctreeLevelInfo(slice_config, level_index)
        self._tiles = {}

    def get_chunk(
        self, row: int, col: int, create_chunks=False
    ) -> Optional[OctreeChunk]:
        """Return the OctreeChunk at this location if it exists.

        If create is True, an OctreeChunk will be created if one does not exist.

        Parameters
        ----------
        row : int
            The row in the level.
        col : int
            The column in the level.
        create_chunks : bool
            If True, create the OctreeChunk if it does not exist.

        Return
        ------
        Optional[OctreeChunk]
            The OctreeChunk if one exists at this location.
        """
        try:
            return self._tiles[(row, col)]
        except KeyError:
            if not create_chunks:
                return None

        # Create a chunk at this location and return it.
        octree_chunk = self._create_chunk(row, col)
        self._tiles[(row, col)] = octree_chunk
        return octree_chunk

    def _create_chunk(self, row: int, col: int) -> OctreeChunk:
        """Create a new OctreeChunk for this location in the level.

        Parameters
        ----------
        row : int
            The row in the level.
        col : int
            The column in the level.

        Return
        ------
        OctreeChunk
            The newly created chunk.
        """
        level_index = self.info.level_index
        location = OctreeLocation(self.slice_id, level_index, row, col)

        scale = self.info.scale
        scale_vec = np.array([scale, scale], dtype=np.float32)

        tile_size = self.info.slice_config.tile_size
        scaled_size = tile_size * scale

        pos = np.array(
            [col * scaled_size, row * scaled_size], dtype=np.float32
        )

        # Geom is used by the visual for rendering this chunk.
        geom = OctreeChunkGeom(pos, scale_vec)

        data = self._get_data(row, col)

        # Return the newly created chunk.
        return OctreeChunk(data, location, geom)

    def _get_data(self, row: int, col: int) -> ArrayLike:

        tile_size = self.info.slice_config.tile_size

        array_slice = (
            slice(row * tile_size, (row + 1) * tile_size),
            slice(col * tile_size, (col + 1) * tile_size),
        )

        if self.data.ndim == 3:
            array_slice += (slice(None),)  # Add the colors.

        data = self.data[array_slice]

        # if not delay_ms.is_zero:
        #    data = _add_delay(data, delay_ms)

        return data



"""VispyTiledImageLayer class.

A tiled image that uses TiledImageVisual and TextureAtlas2D so
adding/removing tiles is extremely fast.
"""
from dataclasses import dataclass
from typing import List

from vispy.scene.visuals import create_visual_node

from ...layers.image.experimental import OctreeChunk
from ...layers.image.experimental.octree_image import OctreeImage
from ...utils.perf import block_timer
from ..vispy_image_layer import VispyImageLayer
from .tile_grid import TileGrid
from .tiled_image_visual import TiledImageVisual

# Create the scene graph Node version of this visual.
TiledImageNode = create_visual_node(TiledImageVisual)


@dataclass
class ChunkStats:
    """Statistics about chunks during the update process."""

    seen: int = 0
    start: int = 0
    deleted: int = 0
    low: int = 0
    created: int = 0
    final: int = 0


class VispyTiledImageLayer(VispyImageLayer):
    """A tiled image drawn using a single TiledImageVisual.

    Tiles are rendered using TiledImageVisual which uses a TextureAtlas2D,
    see those classes for more details.

    History
    -------

    An early tiled visual we had created a new ImageVisual for each tile.
    This led to crashes with PyQt5 due to the constant scene graph changes.
    Also each new ImageVisual caused a slow down, the shader build was one
    reason. Finally rendering was slower because it required a texture swap
    for each tile. This new tiled version solves those problems.

    Parameters
    ----------
    layer : OctreeImage
        The layer we are drawing.

    Attributes
    ----------
    grid : TileGrid
        Optional grid outlining the tiles.
    """

    def __init__(self, layer: OctreeImage):
        # All tiles are stored in a single TileImageVisual.
        visual = TiledImageNode(tile_shape=layer.tile_shape)

        # Pass our TiledImageVisual to the base class, it will become our
        # self.node which is VispyBaseImage.node.
        super().__init__(layer, visual)

        # Optional grid shows tile borders.
        self.grid = TileGrid(self.node)

        self.layer.events.loaded.connect(self._on_loaded)

    @property
    def num_tiles(self) -> int:
        """Return the number of tiles currently being drawn.

        Return
        ------
        int
            The number of tiles currently being drawn.
        """
        return self.node.num_tiles

    def set_data(self, node, data) -> None:
        """Set our image data, not implemented.

        ImageVisual has a set_data() method but we don't. No one can set
        the data for the whole image, that's why it's a tiled image in the
        first place. Instead of set_data() we pull our data one chunk at a
        time by calling self.layer.visible_chunks in our _update_view()
        method.
        """
        raise NotImplementedError()

    def _update_chunks(self) -> ChunkStats:
        """Add or remove tiles to match the chunks which are currently visible.

        1) Remove tiles which are no longer visible.
        2) Create tiles for newly visible chunks.
        3) Optionally update our grid to outline the visible chunks.
        """
        # Get the currently visible chunks from the layer.
        visible_chunks: List[OctreeChunk] = self.layer.visible_chunks

        stats = ChunkStats(seen=len(visible_chunks))

        # Create the visible set of chunks using their keys.
        # TODO_OCTREE: use __hash__ not OctreeChunk.key?
        visible_set = set(octree_chunk.key for octree_chunk in visible_chunks)

        stats.start = self.num_tiles

        # Remove tiles for chunks which are no longer visible.
        self.node.prune_tiles(visible_set)

        stats.low = self.num_tiles
        stats.deleted = stats.start - stats.low

        if self.layer.track_view:
            # Add tiles for visible chunks that do not already have a tile.
            self.node.add_chunks(visible_chunks)

        stats.final = self.num_tiles
        stats.created = stats.final - stats.low

        if self.layer.show_grid:
            self.grid.update_grid(self.node.octree_chunk)
        else:
            self.grid.clear()

        return stats

    def _update_tile_shape(self) -> None:
        """Check if the tile shape was changed on us."""
        # This might be overly dynamic, but for now if we see there's a new
        # tile shape we nuke our texture atlas and start over with the new
        # tile shape.
        #
        # We added this because the QtTestImage GUI sets the tile shape
        # after the layer is created. But the ability might come in handy
        # and it was not hard to implement.
        tile_shape = self.layer.tile_shape
        if self.node.tile_shape != tile_shape:
            self.node.set_tile_shape(tile_shape)

    def _update_view(self):
        if not self.node.visible:
            return

        self._update_tile_shape()  # In case the tile shape changed!

        with block_timer("_update_chunks") as elapsed:
            stats = self._update_chunks()

        if stats.created > 0 or stats.deleted > 0:
            print(
                f"tiles: {stats.start} -> {stats.final} "
                f"create: {stats.created} delete: {stats.deleted} "
                f"time: {elapsed.duration_ms:.3f}ms"
            )

    def _on_camera_move(self, event=None) -> None:
        """Called on any camera movement.

        Update tiles based on which chunks are currently visible.
        """
        super()._on_camera_move()
        self._update_view()

    def _on_loaded(self, _event):
        self._update_view()
